@@ -1,13 +1,13 @@
 package POE::Component::IKC::Responder;
 
 ############################################################
-# $Id: Responder.pm 322 2008-01-16 16:40:45Z fil $
+# $Id: Responder.pm 1248 2014-07-07 09:06:58Z fil $
 # Based on tests/refserver.perl
 # Contributed by Artur Bergman <artur@vogon-solutions.com>
 # Revised for 0.06 by Rocco Caputo <troc@netrus.net>
 # Turned into a module by Philp Gwyn <fil@pied.nu>
 #
-# Copyright 1999-2008 Philip Gwyn.  All rights reserved.
+# Copyright 1999-2014 Philip Gwyn.  All rights reserved.
 # This program is free software; you can redistribute it and/or modify
 # it under the same terms as Perl itself.
 #
@@ -17,16 +17,17 @@ package POE::Component::IKC::Responder;
 use strict;
 use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $ikc);
 use Carp;
-use Data::Dumper;
+use Data::Dump qw( pp );
 
 use POE qw(Session);
 use POE::Component::IKC::Specifier;
+use POE::Component::IKC::Timing;
 use Scalar::Util qw(reftype);
 
 require Exporter;
 @ISA = qw(Exporter);
 @EXPORT = qw(create_ikc_responder $ikc);
-$VERSION = '0.2001';
+$VERSION = '0.2402';
 
 sub DEBUG { 0 }
 
@@ -36,6 +37,7 @@ sub DEBUG { 0 }
 # This is just a convenient way to create only one responder.
 sub create_ikc_responder 
 {
+    carp "create_ikc_responder is deprecated.  Use ", __PACKAGE__, "->spawn instead";
     __PACKAGE__->spawn();
 }
 
@@ -46,10 +48,11 @@ sub spawn
     POE::Session->create( 
         package_states => [
             $package, [qw(
-                      _start _stop
+                      _start _stop _child
                       request post call raw_message post2
-                      remote_error
-                      register unregister default register_local
+                      remote_error channel_error
+                      register unregister register_local register_channel
+                      default
                       publish retract subscribe unsubscribe
                       published
                       monitor inform_monitors shutdown 
@@ -84,6 +87,13 @@ sub _stop
     # warn Denter $poe_kernel;
 }
 
+sub _child
+{
+    my( $heap, $reason, $session, $ret ) = @_[ HEAP, ARG0, ARG1, ARG2 ];
+    DEBUG and 
+        warn "$$: $_[HEAP] responder _child $reason, $session\n";
+}
+
 #----------------------------------------------------
 # Shutdown everything IKC related that we know about
 sub shutdown
@@ -104,8 +114,9 @@ sub request
 # Register foreign kernels so that we can send states to them
 sub register
 {
-    my($heap, $channel, $rid, $aliases) = @_[HEAP, SENDER, ARG0, ARG1];
-    $heap->{self}->register($channel, $rid, $aliases);
+    my($heap, $channel, $rid, $aliases, $pid) = @_[HEAP, SENDER, ARG0..$#_];
+    # warn "pid=$pid" if $pid;
+    $heap->{self}->register($channel, $rid, $aliases, $pid);
 }
 
 #----------------------------------------------------
@@ -133,6 +144,16 @@ sub default
     $heap->{self}->default($name);
 }
 
+#----------------------------------------------------
+# Register a channel.  So we can tell it to shutdown before it finishes
+# negociating
+sub register_channel
+{
+    my($heap, $channel) = @_[HEAP, SENDER];
+    $heap->{self}->register_channel($channel);
+}
+
+
 
 ##############################################################################
 ## This state allows sessions to monitor a remote kernel
@@ -162,8 +183,6 @@ sub post
 sub post2
 {
     my($heap, $to, $sender, $params) = @_[HEAP, ARG0, ARG1, ARG2];
-    # use Data::Dumper;
-    # warn "post2 params=", Dumper $params;
     $heap->{self}->post($to, $params, $sender);
 }
 
@@ -193,10 +212,20 @@ sub raw_message
 sub remote_error
 {
     my($heap, $msg) = @_[HEAP, ARG0];
-
-    warn "$$: Remote error: $msg\n";
+    $heap->{self}->remote_error( $msg );
 }
 
+#----------------------------------------------------
+# Local channel had an error
+sub channel_error
+{
+    my($heap, $msg) = @_[HEAP, ARG0];
+    return $heap->{self}->channel_error( $msg );
+}
+
+
+
+
 ##############################################################################
 # publish/retract/subscribe mechanism of setting up foreign sessions
 
@@ -235,9 +264,9 @@ sub subscribe
     $sessions=[$sessions] unless ref $sessions;
     return unless @$sessions;
 
+    $sender = $sender->ID if ref $sender;
     if($callback and 'CODE' ne ref $callback)
     {
-        $sender = $sender->ID if ref $sender;
         my $state=$callback;
         $callback=sub 
         {
@@ -246,7 +275,7 @@ sub subscribe
             $kernel->post($sender, $state, @_);
         };
     }
-    $heap->{self}->subscribe($sessions, $callback, $sender->ID);
+    $heap->{self}->subscribe($sessions, $callback, $sender);
 }
 
 # Called by a foreign IKC session 
@@ -260,19 +289,21 @@ sub do_you_have
     my $self=$heap->{self};
     
     DEBUG and 
-        warn "Wants to subscribe to ", specifier_name($ses), "\n";
+        warn "$$: Wants to subscribe to ", specifier_name($ses), "\n";
     if(exists $self->{'local'}{$ses->{session}} and
        (not $ses->{state} or 
         exists $self->{'local'}{$ses->{session}}{$ses->{state}}
        ))
     {
         $ses->{kernel}||=$kernel->ID;       # make sure we uniquely identify 
-        DEBUG and warn "Allowed (we are $ses->{kernel})\n";
+        DEBUG and 
+            warn "$$: Allowed (we are $ses->{kernel})\n";
         return [$ses, $kernel->ID];         # this session
     } else
     {
-        DEBUG and warn specifier_name($ses), " is not published in this kernel\n";
-        return "NOT ".specifier_name($ses);
+        DEBUG and 
+            warn "$$: ", specifier_name($ses), " is not published in this kernel\n";
+        return "Refused subscription ".specifier_name($ses);
     }
 }
 
@@ -327,7 +358,7 @@ use POE::Component::IKC::Proxy;
 use POE::Component::IKC::LocalKernel;
 use POE qw(Session);
 
-use Data::Dumper;
+use Data::Dump qw( pp );
 
 sub DEBUG { 0 }
 sub DEBUG2 { 0 }
@@ -346,6 +377,7 @@ sub new
             rsvp=>{},
             kernel=>{},
             channel=>{},
+            channel_startup=>{},
             default=>{},
             monitors=>{},
             poe_kernel=>$kernel,
@@ -362,11 +394,23 @@ sub shutdown
         warn "$$: Some one wants us to go away... off we go\n";
     # kill our alias
     $kernel->alias_remove('IKC');
+
     # tell every channel to shutdown
     while(my($rid, $c)=each %{$self->{channel}}) {
-        DEBUG and warn "$$: Posting shutdown to $rid (id=$c)\n";
+        DEBUG and 
+                warn "$$: Posting shutdown to $rid (id=$c)\n";
         $kernel->post($c, 'shutdown');
     }
+    $self->{channel} = {};
+
+    # even the channels that haven't negociated yet
+    foreach my $c ( keys %{ $self->{channel_startup} } ) {
+        DEBUG and 
+                warn "$$: Posting shutdown to channel (id=$c)\n";
+        $kernel->post( $c, 'shutdown' );
+    }
+    $self->{channel_startup} = {};
+
     # tell monitors to shutdown
     $self->inform_monitors('*', 'shutdown');
     # kill pending subscription states
@@ -377,6 +421,54 @@ sub shutdown
 #    warn Dump $kernel;
 }
 
+
+#----------------------------------------------------
+# Error in a remote kernel
+sub remote_error
+{
+    my( $self, $msg ) = @_;
+    $self->_do_error( remote => $msg );
+}
+
+# Error in the local kernel
+sub local_error
+{
+    my( $self, $msg ) = @_;
+    $self->_do_error( local => $msg );
+}
+
+# Error in a local channel
+sub channel_error
+{
+    my( $self, $msg ) = @_;
+    return $self->_do_error( channel => $msg, 1 );
+}
+
+sub _do_error
+{
+    my( $self, $where, $msg, $ignore ) = @_;
+    my $n;
+    eval {
+        my $kernel = '*';
+        my $when = $where;
+        if( ref $msg ) {
+            ( $msg, $kernel, $when ) = @$msg;
+            $kernel ||= '*';
+            $when = "$where-$when" unless $when =~ /^$where-/;
+        }
+        $n = $self->inform_monitors( $kernel, 'error', $when, $msg );
+    };
+    return $n if $n;
+    warn "$$: \u$where error: ", pp( $msg ), "\n" unless $ignore;
+    return;
+}
+
+
+
+
+
+
+
 #----------------------------------------------------
 # Foreign kernel called something here
 sub request    
@@ -384,50 +476,22 @@ sub request
     my($self, $request)=@_;;
     my($kernel)=@{$self}{qw(poe_kernel)};
     DEBUG2 and 
-        warn "IKC request=", Dumper $request;
+        warn "IKC request=", pp $request;
 
     # We ignore the kernel for now, but we should really use it to decide
     # weither we should run the request or not
+    my $when = 'request';
     my $to=specifier_parse($request->{event});
-
+    my $rkernel = $to->{kernel};
     eval {
         die "$request->{event} isn't a valid specifier" unless $to;
-        my $args=$request->{params};
-
-        ### allow proxied states to have multiple ARGs
-        if($to->{state} eq 'IKC:proxy') {
-            $to->{state}=$args->[0];
-            $args=$args->[1];
-            DEBUG and warn "IKC proxied request for ", specifier_name($to), "\n";
-        } 
-        else {
-            DEBUG and warn "IKC request for ", specifier_name($to), "\n";
-            $args=[$args];
-        }
+        my $args = $self->_req_args( $request, $to );
 
         # this is where we'd catch a disconnect message
         # 2001/07 : eh?
 
-
-        # find out if the state we want to get at has been published
-        if(exists $self->{rsvp}{$to->{session}} and
-           exists $self->{rsvp}{$to->{session}}{$to->{state}} and
-           $self->{rsvp}{$to->{session}}{$to->{state}}
-          ) {
-            $self->{rsvp}{$to->{session}}{$to->{state}}--;
-            DEBUG and warn "Allow $to->{session}/$to->{state} is now $self->{rsvp}{$to->{session}}{$to->{state}}\n";
-        }
-        elsif(not exists $self->{'local'}{$to->{session}}) {
-            my $p=$self->published;
-            die "Session '$to->{session}' is not available for remote kernels:",
-                join "\n", '',
-                    map({ "    $_=>[" . join(', ', @{$p->{$_}}) . "]"} keys %$p),
-                    '';
-        }
-        elsif(not exists $self->{'local'}{$to->{session}}{$to->{state}}) {
-            die "Session '$to->{session}' has not published state '",
-                $to->{state}, "'\n";
-        }
+        $when = 'check';
+        $self->_req_is_published( $to );
 
         # maybe caller specified #arg?  This got into $msg->{rsvp}, which
         # went to the remote side, then came back here as $to
@@ -435,10 +499,12 @@ sub request
             push @$args, $to->{args};   # it goes on the end
         }
 
+        $when = 'resolve';
         my $session=$kernel->alias_resolve($to->{session});
-        die "Unknown session '$to->{session}'\n" unless $session;
-        # warn "No FROM" unless $request->{from};
-        _thunked_post($request->{rsvp}, [$session, $to->{state}, @$args],
+        die "Unknown session '$to->{session}'" unless $session;
+
+        $when = 'invocation';
+        _thunked_post($request->{rsvp}, ["$session", $to->{state}, @$args],
                           $request->{from}, $request->{wantarray});
     };
 
@@ -446,41 +512,101 @@ sub request
     # Error handling consists of posting a "remote_error" state to
     # the foreign kernel.
     # $request->{errors_to} is set by the local IKC::Channel
-    if($@)
-    {
-        chomp($@);
-        my $err=$@.' ['.specifier_name($to).']';
-        $err.=' sent by ['.specifier_name($request->{from}).']'
-                                                    if $request->{from};
-        warn "$err\n";
-        DEBUG && warn "$$: Error in request: $err\n";
-        unless($request->{is_error})    # don't send an error message back
-        {                               # if this was an error itself
-            $self->send_msg({ event=>$request->{errors_to},
-                              params=>$err, is_error=>1,
-                            });
-        } 
-        else {
-            warn $$,  Dumper $request;
-        }
+    if($@) {
+        $self->_error_response( $@, $request, $to, $rkernel, $when );
     }
 }
 
+sub _req_args
+{
+    my( $self, $request, $to ) = @_;
+    my $args = $request->{params};
+
+    ### allow proxied states to have multiple ARGs
+    if($to->{state} eq 'IKC:proxy') {
+        $to->{state}=$args->[0];
+        $args=$args->[1];
+        DEBUG and warn "IKC proxied request for ", specifier_name($to), "\n";
+    } 
+    else {
+        DEBUG and warn "IKC request for ", specifier_name($to), "\n";
+        $args=[$args];
+    }
+    return $args;
+}
+
+sub _req_is_published
+{
+    my( $self, $to ) = @_;
+
+    # find out if the state we want to get at has been published
+    if(exists $self->{rsvp}{$to->{session}} and
+       exists $self->{rsvp}{$to->{session}}{$to->{state}} and
+       $self->{rsvp}{$to->{session}}{$to->{state}}
+      ) {
+        $self->{rsvp}{$to->{session}}{$to->{state}}--;
+        DEBUG and warn "Allow $to->{session}/$to->{state} is now $self->{rsvp}{$to->{session}}{$to->{state}}\n";
+    }
+    elsif(not exists $self->{'local'}{$to->{session}}) {
+        my $p=$self->published;
+        DEBUG and
+            warn "$$: Available: ",
+                join "\n", '',
+                    map({ "    $_=>[" . join(', ', @{$p->{$_}}) . "]"} keys %$p),
+                    '';
+        die "Session '$to->{session}' is not available for remote kernels\n",
+    }
+    elsif(not exists $self->{'local'}{$to->{session}}{$to->{state}}) {
+        die "Session '$to->{session}' has not published state '",
+            $to->{state}, "'\n";
+    }
+    return 1;
+}
+
+sub _error_response
+{
+    my( $self, $err, $request, $to, $rkernel, $when ) = @_;
+
+    chomp( $err );
+    $err=$err.'.  Request '.specifier_name($to);
+    $err.=' sent by '.specifier_name($request->{from})
+                                                    if $request->{from};
+
+    # Tell local sessions about this error
+    $self->local_error( [ $err, $rkernel, $when ] );
+
+    # never respond to an error message with an error message
+    return if $request->{is_error};
+
+    # Tell remote sessions about this error
+    $self->send_msg( { event=>$request->{errors_to},
+                       params=>[ $err, $rkernel, $when ],
+                       is_error=>1,
+                   } );
+}
+
+
+
+
 #----------------------------------------------------
 # Register foreign kernels so that we can send states to them
 sub register
 {
-    my($self, $channel, $rid, $aliases)=@_;
+    my($self, $channel, $rid, $aliases, $pid)=@_;
     $aliases=[$aliases] if not ref $aliases;
 
     my($kernel)=@{$self}{qw(poe_kernel)};
     
+    $channel = $channel->ID;
+    delete $self->{channel_startup}{ $channel };
+
     if($self->{channel}{$rid}) {
         warn "$$: Remote kernel '$rid' already exists\n";
         return;
     } 
     else {
-        DEBUG and warn "$$: Registered remote kernel '$rid'\n";
+        DEBUG and 
+            warn "$$: Registered remote kernel '$rid' (id=$channel)\n";
         $self->{channel}{$rid}=$channel;
         $self->{remote}{$rid}=[];       # list of proxy sessions
         $self->{alias}{$rid}=$aliases;  
@@ -501,7 +627,8 @@ sub register
         $self->{kernel}{$name}=$rid;    # find real remote ID
         $self->{remote}{$name}||=[];    # list of proxy sessions
     }
-    $self->inform_monitors($rid, 'register');
+    # warn "pid=$pid" if $pid;
+    $self->inform_monitors($rid, 'register', $pid);
 
     return 1;
 }
@@ -526,8 +653,6 @@ sub register_local
     $self->{remote}{$rid}||=[];       # list of proxy sessions
     $self->{alias}{$rid}||=[];
 
-    # use Data::Dumper;
-    # die Dumper $aliases;
     foreach my $name (@$aliases) {
         unless(defined $name) {
             DEBUG and warn "$$: attempt to register undefined local kernel alias\n";
@@ -545,6 +670,19 @@ sub register_local
     return 1;
 }
 
+#----------------------------------------------------
+# Register a starting channel
+sub register_channel
+{
+    my( $self, $channel ) = @_;
+    $channel = $channel->ID;
+    DEBUG and 
+        warn "$$: Registered channel (id=$channel)\n";
+    $self->{channel_startup}{ $channel } = 1;
+    return;
+}
+
+
 #----------------------------------------------------
 sub default
 {
@@ -572,6 +710,7 @@ sub unregister
     my($self, $channel, $rid, $aliases)=@_;
     my($kernel)=@{$self}{qw(poe_kernel)};
     return unless $rid;
+    $channel = $channel->ID;
 
     unless($aliases) {
         unless($self->{channel}{$rid}) {    # unregister one alias only
@@ -644,18 +783,18 @@ sub send_msg
 
     my $to=specifier_parse($msg->{event});
     unless($to) {
-        die "Bad state ", Dumper $msg;
+        die "Bad state ", pp $msg;
     }
     unless($to) {
-        warn "Bad or missing 'event' parameter '$msg->{event}' to poe://IKC/$e\n";
+        warn "Bad or missing 'event' parameter '$msg->{event}' to IKC/$e\n";
         return;
     }
     unless($to->{session}) {
-        warn "Need a session name in poe://IKC/$e\n";
+        warn "Need a session name for IKC/$e\n", pp $to;
         return;
     }
     unless($to->{state})   {
-        warn "Need a state name in poe://IKC/$e\n", Dumper $to;
+        warn "Need a state name for IKC/$e\n", pp $to;
         return;
     }
 
@@ -674,9 +813,10 @@ sub send_msg
         # if you want
         $self->{poe_kernel}->can('alias_list')) 
     {
-        my $sid=$sender->ID if ref $sender;
+        my $sid=$sender;
+        $sid = $sender->ID if ref $sender;
         foreach my $a ($self->{poe_kernel}->alias_list($sender)) {
-            $sid.=" ($a)";
+            $sid .= " ($a)";
             if($self->{'local'}{$a}) {  # SENDER published something
                 $msg->{from}={  kernel=>$self->{poe_kernel}->ID,
                                 session=>$a,
@@ -702,13 +842,11 @@ sub send_msg
 
     # Get a list of channels to send the message to
     my @channels=$self->channel_list( $name );
-    unless(@channels)
-    {
-        warn "$$: MSG TO ", Dumper $to;
-        warn (($name eq '*')
-                  ? "$$: Not connected to any foreign kernels.\n"
-                  : "$$: Unknown kernel '$name'.\n");
-        warn "$$: Known kernels: ". $self->channel_names;
+    unless(@channels) {
+        my $err = (($name eq '*')
+                    ? "$$: Not connected to any foreign kernels."
+                    : "$$: Unknown kernel '$name'.");
+        $self->inform_monitors( '*', 'error', 'resolve', $err );
         return 0;
     }
 
@@ -871,9 +1009,6 @@ sub post
     my($self, $to, $params, $sender) = @_;
 
     $to="poe://$to" unless ref $to or $to=~/^poe:/;
-    # use Data::Dumper;
-    # warn "params=", Dumper $params;
-
     $self->send_msg({params=>$params, 'event'=>$to}, $sender);
 }
 
@@ -888,19 +1023,19 @@ sub call
     $rsvp="poe://$rsvp" if $rsvp and not ref $rsvp and $rsvp!~/^poe:/;
 
     unless($rsvp) {
-        warn "$$: Missing 'rsvp' parameter in poe://IKC/call\n";
+        warn "$$: Missing 'rsvp' parameter in poe:IKC/call\n";
         return;
     }
     my $t=specifier_parse($rsvp);
     unless($t) {
-        warn "$$: Bad 'rsvp' parameter '$rsvp' in poe://IKC/call\n";
+        warn "$$: Bad 'rsvp' parameter '$rsvp' in poe:IKC/call\n";
         return;
     }
     $rsvp=$t;
     unless($rsvp->{state})
     {
-        DEBUG and warn Dumper $rsvp;
-        warn "$$: rsvp state not set in poe://IKC/call\n";
+        DEBUG and warn pp $rsvp;
+        warn "$$: rsvp state not set in poe:IKC/call\n";
         return;
     }
 
@@ -913,8 +1048,6 @@ sub call
     }
     DEBUG2 and warn "RSVP is ", specifier_name($rsvp), "\n";
 
-    # use Data::Dumper;
-    # warn "params=", Dumper $params;
     $self->send_msg({params=>$params, 'event'=>$to,
                      rsvp=>$rsvp
                     }, $sender
@@ -957,7 +1090,7 @@ sub publish
         die "\$states isn't an array ref" unless ref($states) eq 'ARRAY';
         foreach my $q (@$states) {
             DEBUG and 
-                print STDERR "Published poe:$alias/$q\n";
+                print STDERR "$$: Published poe:$alias/$q\n";
             $p->{$q}=1;
         }
     }
@@ -1059,8 +1192,14 @@ sub subscribe
                      rsvp=>{kernel=>$id, session=>'IKC', state=>$unique.$spec},
                     }, 
                  );
-            # TODO What if this post failed?  Session that posted this would
-            # surely want to know
+            DEBUG and warn "$$: do_you_have sent to $count sessions\n";
+            if( $count == 0 ) {
+                # This post failed.  Session that posted this would
+                # surely want to know
+                $self->inform_monitors( '*', 'error', 
+                                        'subscribe', 
+                                        "Unknown kernel $ses->{kernel}" );
+            }
         } else
         {                       # Bleh.  User shouldn't be that dumb       
             die "You can't subscribe to a session within the current kernel.";
@@ -1128,8 +1267,11 @@ sub _subscribe_receipt
     my $del;
 
     if(not $ses or not ref $ses) {               # REFUSED
-        warn "$$: Refused to subscribe to $spec";
-        warn "$$: $resp" if $resp;
+        $resp ||= "Refused subscription to $spec";
+        $self->inform_monitors( '*', 'error', 
+                                         'subscribe', 
+                                         $resp )
+            or warn "$$: $resp";
         $accepted=0;
         $del=$unique.$spec;
     } 
@@ -1251,7 +1393,7 @@ sub monitor
 
 #----------------------------------------------------
 # Tell monitors about something in foreign kernel
-# $rid == kernel name (in which case we ALSO inform about alises) or alias
+# $rid == kernel name (in which case we ALSO inform about aliases) or alias
 #         or * (tell every monitor about something... future use)
 # $event == name of event we are informing about
 # @params == other stuff
@@ -1266,7 +1408,12 @@ sub inform_monitors
 
     my $real=1 if $self->{channel}{$rid};
     DEBUGM and 
-        warn "$$: $rid is", ($real ? '' : "n't"), " real\n";
+        do {
+            warn "$$: inform $event $rid\n";
+            warn "$$: $rid is", ($real ? '' : "n't"), " real\n";
+        };
+
+    my $count = 0;
 
     # got to be a better way of doing this...
     my @todo=($rid);
@@ -1296,15 +1443,17 @@ sub inform_monitors
                 # ARG4.... = per-message info
             $kernel->post($sender, $e, $states->{__name}, $rid, $real,
                             $states->{data}, @params);
+            $count++;
         }
     }
 
     # $rid might be an alias to something else, inform about those as well
     if($self->{channel}{$rid}) {
         foreach my $ra (@{$self->{alias}{$rid}}) {
-            $self->inform_monitors($ra, $event, @params);
+            $count += $self->inform_monitors($ra, $event, @params);
         }
     }
+    return $count;
 }
 
 
@@ -1313,14 +1462,12 @@ sub inform_monitors
 # kernel.  Currently, the thunks are used as a "proof of concept" and
 # to accur extra over head. :)
 # 
-# The original idea was to make sure that $_[SENDER] would be something
-# valid to the posted state.  However, garbage collection becomes a problem
-# If we are to allow the poste session to "hold on to" $_[SENDER]. 
-#
-# Having the Responder save all Thunks until the related foreign connection
-# disapears would be wasteful.  Or, POE could use soft-refs to track
-# active sessions and the thunk would "kill" itself in _stop.  This
-# would force POE to require 5.005, however.
+# On the first request, a thunk is created.  It is kept alive with an alias.
+# On the next request, we check to see if the extref_count is zero.  If it
+# is, we reuse the same request.  If not, we create a new thunk and continue
+# using that.  What's more, we tell the thunk that it is active, so it should
+# clear its alias.  This way, when the user code decrements the extref_count
+# back to zero, the thunk they reserved can be cleared.
 #
 
 # Export thunk the quick way.
@@ -1328,70 +1475,160 @@ sub inform_monitors
 package POE::Component::IKC::Responder::Thunk;
 
 use strict;
-use POE qw(Session);
-use Data::Dumper;
+
+use Carp;
+use Data::Dump qw( pp );
+use POE::Component::IKC;
+use POE::Session;
+use POE;
 
 sub DEBUG { 0 }
+sub DEBUG2 { 0 }
 
+#----------------------------------------------------
 {
-    my $name=__PACKAGE__.'00000000';
-    $name=~s/\W//g;
+    my $NAME=__PACKAGE__.'00000000';
+    $NAME=~s/\W+//g;
+    my $current_thunk;
+
+    #------------------------------
     sub thunk
     {
-        # my($rsvp, $args, $from, $wantarray)=@_;
-        POE::Session->create( 
-                package_states => [__PACKAGE__, [qw(_start _stop _default)]],
-                args => [$name++, @_]
+        # my($rsvp, $call, $from, $wantarray)=@_;
+        unless( __active_thunk() ) {
+            __create_thunk();
+        }
+
+        # we use call to make sure no other call to us could
+        # happen between _start and __thunk
+        $poe_kernel->call( $current_thunk => '__thunk', @_ );
+    }
+
+    #------------------------------
+    sub __create_thunk
+    {
+        my $thunk = 
+            POE::Session->create( 
+                package_states => [ __PACKAGE__, 
+                                    [ qw(_start _stop _default
+                                        __thunk __active
+                                    ) ]
+                                  ],
+                args => [++$NAME]
             );
+
+        $current_thunk = $thunk->ID;
+    }
+
+    #------------------------------
+    sub __active_thunk 
+    {
+        return unless $current_thunk;
+        # 2009/05 - These next 2 lines call undocumented internal methods of 
+        # the kernel.  If the kernel changes, they will break.
+        # If they break, please contact gwyn-at-cpan.org.
+        # 2011/08 - These have been changed for 1.311
+        my $count = _ref_count( $current_thunk );
+        if( defined $count ) {
+            DEBUG and 
+                warn "$$: $NAME count=$count\n";
+            if( 0==$count ) {
+                DEBUG and 
+                    warn "$$: $NAME reuse\n";
+                return 1;
+            }
+            DEBUG and 
+                warn "$$: new thunk\n";
+            $poe_kernel->call( $current_thunk => '__active' );
+        }
+        undef( $current_thunk );
+        return;
     }
 }
 
+sub _ref_count
+{
+    my( $id ) = @_;
+    # This code is badly behaved!
+    return unless $poe_kernel->_data_ses_exists( $id );
+    if( $poe_kernel->can( '_data_extref_count_ses' ) ) {
+        return $poe_kernel->_data_extref_count_ses( $id )||0;
+    }
+    else {
+        # This is for the code that dngor had that extrefs as a sub object, not
+        # as a mixin'
+        $poe_kernel->[ POE::Kernel::KR_EXTRA_REFS() ]->count_session_refs( $id )||0;
+    }
+}
+
+
+#----------------------------------------------------
 sub _start
 {
-    my($kernel, $heap,          $name, $rsvp,  $args, $from, $wantarray)=
-                @_[KERNEL, HEAP, ARG0, ARG1,    ARG2,  ARG3,  ARG4];
-    $heap->{from}=$from;
-    $heap->{name}=$name;
+    my($kernel, $heap, $name )= @_[KERNEL, HEAP, ARG0];
+    $heap->{alias} = $heap->{name} = $name;
+    DEBUG and warn "$$: $name create\n";
+    $kernel->alias_set( $heap->{alias} );
+}
+
+#----------------------------------------------------
+sub __active
+{
+    my($kernel, $heap) = @_[KERNEL, HEAP];
+    DEBUG and 
+        warn "$$: $heap->{name} active\n";
+    $kernel->alias_set( delete $heap->{alias} ) if $heap->{alias};
+    return 1;
+}
+
+#----------------------------------------------------
+sub _stop
+{
+    DEBUG and 
+        warn "$$: $_[HEAP]->{name} delete\n";
+}
+
+#----------------------------------------------------
+sub __thunk
+{
+    my($kernel, $heap,          $rsvp, $call, $from, $wantarray)=
+                @_[KERNEL, HEAP, ARG0, ARG1,  ARG2,  ARG3];
+    $heap->{from} = $from;
     # warn "no FROM" unless $from;
 
-    DEBUG and warn "$name created\n";
-
     if($rsvp) {                         # foreign session wants returned value
+        DEBUG2 and warn "Calling ", pp $call;
 
-        DEBUG and do { warn "Wants an array" if $wantarray};
+        DEBUG2 and do { warn "Wants an array" if $wantarray};
 
         my(@ret, $yes);
         if($wantarray) {
-            @ret=$kernel->call(@$args);
+            @ret=$kernel->call(@$call);
             $yes = 0<@ret;
         } else {
-            $ret[0]=$kernel->call(@$args);
+            $ret[0]=$kernel->call(@$call);
             $yes = defined $ret[0];
         }
         if($yes) {
-            DEBUG and do {
+            DEBUG2 and do {
                 local $"=', ';
-                warn "Posted response '@ret' to ", Dumper $rsvp;
+                warn "Posted response '@ret' to ", pp $rsvp;
             };
             # This is the POSTBACK
             $POE::Component::IKC::Responder::ikc->send_msg(
                     {params=>($wantarray ? \@ret : $ret[0]), event=>$rsvp},
-                    $args->[0]);
+                    $call->[0]);
         }
     }
     else {
-        DEBUG and do {
-            warn "Posting ", Dumper $args;
-        };
-        $kernel->post(@$args);
+        # 2009/05 - use ->call() so that {from} can't be modified
+        # before refcount_increment is called
+        DEBUG2 and warn "Posting ", pp $call;
+        $kernel->call(@$call);
     }
 }
 
-sub _stop
-{
-    DEBUG and warn "$_[HEAP]->{name} delete\n";
-}
-
+#----------------------------------------------------
 sub _default
 {
     my($kernel, $heap, $sender, $state, $args)=
@@ -1403,9 +1640,19 @@ sub _default
         return;
     }
 
-    $POE::Component::IKC::Responder::ikc->send_msg(
+    if( not $heap->{from}{state} ) {
+        my $event = { %{$heap->{from}} };
+        $event->{state} = $state;
+
+        $POE::Component::IKC::Responder::ikc->send_msg(
+                {params=>$args, event=>$event}, $sender
+              );
+    }
+    else {
+        $POE::Component::IKC::Responder::ikc->send_msg(
                 {params=>[$state, $args], event=>$heap->{from}}, $sender
               );
+    }
 }
 
 1;
@@ -1427,7 +1674,7 @@ POE::Component::IKC::Responder - POE IKC state handler
 
 =head1 DESCRIPTION
 
-This module implements an POE IKC state handling.  The responder handles
+This module implements POE IKC state handling.  The responder handles
 posting states to foreign kernels and calling states in the local kernel at
 the request of foreign kernels.
 
@@ -1478,7 +1725,7 @@ This logs an state with a hypothetical logger.
 
 =back
 
-See the L</Note about SENDER> below.
+See the L</PROXY SENDER> below.
 
 
 =head2 C<call>
@@ -1539,7 +1786,7 @@ The C<rsvp> state does not need to be published.  IKC keeps track of the
 rsvp state and will allow the foreign kernel to post to it.
 
 
-See the L<Note about SENDER> below.
+See the L<PROXY SENDER> below.
 
 
 
@@ -1810,6 +2057,12 @@ The following states can be monitored:
 
 =over 6
 
+=item C<channel>
+
+Called when a channel becomes ready or goes away.  ARG3 is either C<ready>
+or C<close>.  ARG4 is the numerical ID of the channel's session.  See
+L</CHANNELS> below.
+
 =item C<register>
 
 Called when a remote kernel or alias is registered.  This is equivalent to
@@ -1836,6 +2089,11 @@ You are informed whenever someone tries to do a sane shutdown of IKC and all
 peripheral sessions.  This will called only once, after somebody posts an
 IKC/shutdown event.
 
+=item C<error>
+
+You are informed of errors in local and remote kernels.  ARG3 is the operation that
+failed. ARG4 is the error message.  See L</ERRORS> below.
+
 =item C<data>
 
 Little bit of data (can be scalar or reference) that is passed to the
@@ -1877,6 +2135,9 @@ Most of the time, ARG0 and ARG1 will be the same.  Exceptions are if you are
 monitoring C<*> or if you supplied a full IKC event specifier to
 IKC/monitor rather then just a plain kernel name. 
 
+
+
+
 =head2 Short note about monitoring all kernels with C<*>
 
 There are 2 reasons circonstances in which you will be monitoring all remote
@@ -1925,9 +2186,140 @@ Session is no longer monitoring all kernels, only 'Pulse'.
 Now we aren't even interested in 'Pulse';
 
 
+=head1 CHANNELS
+
+Previous versions of IKC did not adequately allow you to control a connection.
+With 0.2400 we added a much needed feature.
+
+Each connection to a remote kernel is handled by a channel session.  You
+find out the session's ID by monitoring for L</channel> operations.  You may
+close a channel and the corresponding connection to the remote kernel by
+sending it a L</shutdown> event.
+
+    sub _start {
+        # set up the monitor
+        $poe_kernel->call( IKC => monitor => '*' => { channel => 'channel' } );
+    }
+
+    sub channel {
+        my( $self, $rid, $rkernel, $real, $data, $op, $channel ) = @_[ OBJECT, ARG0..$#_ ];
+        return unless $real;    # only care about the real kernel ID
+        if( $op eq 'ready' ) {  # new channel is ready
+            $self->{channel}{ $rkernel } = $channel;
+        }
+        elsif( $op eq 'close' ) {   # channel is gone
+            delete $self->{channel}{ $rkernel };
+        }
+    }
+
+    # this an event posted from your controler logic
+    sub close_channel {
+        my( $self, $rkernel ) = @_[ OBJECT, ARG0 ];
+        # tell the channel to close
+        $poe_kernel->post( $self->{channel}{ $rkernel } => 'shutdown' );
+    }
 
 
 
+=head1 ERRORS
+
+Previous versions of IKC did not adequately allow you to monitor for errors
+on a connection.  With 0.2400 we started monitoring errors.
+
+There are 2 step during which you can have errors: when opening the connection and
+during message exchange.  These 2 steps are handled diffrently.
+
+You use L<POE::Component::IKC::Client/on_error> and
+L<POE::Component::IKC::Server/on_error> to receive errors while a connection
+is being opened.  Note that this includes the initial IKC handshake.
+
+    sub on_error 
+    {
+        my( $op, $errnum, $errstr ) = @_;
+        # Handle this like you would any POE socket error
+        # But remember you can't rely on your session being active
+    }
+
+
+You use L</monitor> on error to receive errors during message exchange.  ARG3 is the
+name of the operation.  ARG4 is the error message.  Current operations are:
+
+=over 4
+
+=item remote-request
+
+Remote kernel was unable to parse a request that was sent from the local kernel.
+
+=item remote-check
+
+Remote kernel has not published an event that was sent from the local kernel.
+
+=item remote-resolve
+
+Remote kernel could not find a session that could handle the request.
+
+=item remote-invocation
+
+Remote kernel had an error when it tried to invoke the request handler. 
+Please note this will not catch errors in the request handler, but only errors
+in the thunk.
+
+
+=item local-request
+
+=item local-check
+
+=item local-resolve
+
+=item local-invocation
+
+These 4 operations are the local equivalent of the previous 4.  They are
+intented for logging.  In general no actions are required.
+
+Note that 'local' and 'remote' refer to where the operation happened, not
+where the request originated.  As an example, kernel A sends a
+poe://B/foo/bar request to kernel B.  Kernel B has not published that event. 
+Monitors on kernel A will see L<remote-check>.  Monitors on kernel B will
+see L<local-check>.
+
+=item channel-error
+
+Receive channel errors during message exchange.   Channel errors are
+equivalent to POE wheel errors.  The message will be C<"[$errnum] $errstr">.
+
+=item subscribe
+
+Failure to subscribe to a remote session.
+
+=item fork
+
+L<POE::Component::IKC::Server> failed to fork.
+
+=item resolve
+
+Error when trying to find a remote kernel or session.
+
+=back
+
+Example monitor for error events:
+
+    sub monitor_error
+    {   
+        my( $self, $rid, $kernel, $real, $data, $op, $message ) = 
+                @_[ OBJECT, ARG0 ... $#_ ];
+        if( $op =~ /^channel-/ and $message =~ /\[(\d+)\] (.*)/ ) {
+            return unless $real;
+            my( $errnum, $errstr ) = ( $1, $2 );
+            if( $op eq 'channel-read' and $errnum == 0 ) {
+                warn "Connection closed";
+                return;
+            }
+        }
+        warn "Error during $op: $message";
+    }
+
+In particular, you will note we don't do anything when we detect the channel
+closed.  Instead, it is recommended to attempt reconnection in the L</unregister> event.
 
 
 
@@ -1935,36 +2327,46 @@ Now we aren't even interested in 'Pulse';
 
 =head2 C<create_ikc_responder>
 
-This function creates the Responder session and object.  However, you don't
-need to call this directly, because L<POE::Component::IKC::Client> or 
-L<POE::Component::IKC::Server> does
-this for you.
+DEPRECATED.  Please use 
+    
+    POE::Compontent::IKC::Responder->spawn();
 
-Deprecated, use L</spawn>.
 
-=head1 L<Note about SENDER>
 
-An attempt is made to provide a sane SENDER param to called or posted
-states.  If the calling session has published some states, SENDER is usable
-during the called state, but not afterwards.  Don't try keeping a reference
-to this session.  This makes callbacks a tad easier.
+=head1 PROXY SENDER
+
+Event handlers invoked via IKC will have a proxy SENDER session. You may use
+it to post back to the remote session.   
+
+    $poe_kernel->post( $_[SENDER], 'response', @args );
+
+Normally this proxy session is available during the invocation of the event
+handler.  You may claim it for longer by setting an external reference:
+
+    $heap->{remote} = $_[SENDER]->ID;
+    $poe_kernel->refcount_increment( $heap->{remote}, 'MINE' );
+
+POE::Component::IKC will detect this and create a new proxy session for future
+calls.  It will then be UP TO YOU to free the session:
+
+    $poe_kernel->refcount_decrement( $heap->{remote}, 'MINE' );
+
+Note that you will have to publish any events that will be posted back.
 
-Furthur, if you IKC/call a remote state, the SENDER during the callback will
-point back to the remote session.
 
 =head1 BUGS
 
-Sending session references and coderefs to a foreign kernel is a bad idea :)  At some
-point it would be desirable to recurse through the paramerters and and turn
-any session references into state specifiers.
+Sending session references and coderefs to a foreign kernel is a bad idea.
+At some point it would be desirable to recurse through the paramerters and
+and turn any session references into state specifiers.
 
-C<rsvp> state in call is a bit problematic.  IKC allows it to be posted to
-once, but doesn't check to see if the foreign kernel is the right one.
+The C<rsvp> state in call is a bit problematic.  IKC allows it to be posted
+to once, but doesn't check to see if the foreign kernel is the right one.
 
 C<retract> does not currently tell foreign kernels that have subscribed to a
-session/state that it has been retracted.
+session/state about the retraction.
 
-C<call()ing> a state in a proxied foreign session doesn't work, for obvious
+C<call()>ing a state in a proxied foreign session doesn't work, for obvious
 reasons.
 
 
@@ -1975,7 +2377,7 @@ Philip Gwyn, <perl-ikc at pied.nu>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright 1999-2008 by Philip Gwyn.  All rights reserved.
+Copyright 1999-2014 by Philip Gwyn.  All rights reserved.
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself.
