@@ -1,13 +1,13 @@
 package POE::Component::IKC::Server;
 
 ############################################################
-# $Id: Server.pm 322 2008-01-16 16:40:45Z fil $
+# $Id: Server.pm 1247 2014-07-07 09:06:34Z fil $
 # Based on refserver.perl and preforkedserver.perl
 # Contributed by Artur Bergman <artur@vogon-solutions.com>
 # Revised for 0.06 by Rocco Caputo <troc@netrus.net>
 # Turned into a module by Philp Gwyn <fil@pied.nu>
 #
-# Copyright 1999-2008 Philip Gwyn.  All rights reserved.
+# Copyright 1999-2014 Philip Gwyn.  All rights reserved.
 # This program is free software; you can redistribute it and/or modify
 # it under the same terms as Perl itself.
 #
@@ -17,25 +17,28 @@ package POE::Component::IKC::Server;
 use strict;
 use Socket;
 use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
+use Carp;
 use POE qw(Wheel::ListenAccept Wheel::SocketFactory);
 use POE::Component::IKC::Channel;
 use POE::Component::IKC::Responder;
+use POE::Component::IKC::Util;
 use POSIX qw(:errno_h);
 use POSIX qw(ECHILD EAGAIN WNOHANG);
 
+
 require Exporter;
 
 @ISA = qw(Exporter);
 @EXPORT = qw(create_ikc_server);
-$VERSION = '0.2001';
+$VERSION = '0.2402';
 
 sub DEBUG { 0 }
 sub DEBUG_USR2 { 1 }
 BEGIN {
     # http://support.microsoft.com/support/kb/articles/Q150/5/37.asp
-    eval '*WSAEAFNOSUPPORT = sub { 10047};';
-    if($^O eq 'MSWin32') {
-        eval '*EADDRINUSE      = sub { 10048 };';
+    eval '*WSAEAFNOSUPPORT = sub { 10047 };';
+    if( $^O eq 'MSWin32' and not eval "EADDRINUSE" ) {
+        eval '*EADDRINUSE  = sub { 10048 };';
     }
 }
 
@@ -45,22 +48,25 @@ BEGIN {
 # This is just a convenient way to create servers.  To be useful in
 # multi-server situations, it probably should accept a bind address
 # and port.
-sub create_ikc_server
+sub spawn
 {
-    my(%params)=@_;
-    $params{package}||=__PACKAGE__;
+    my($package, %params)=@_;
+    $params{package} ||= $package;
 
     unless($params{unix}) {
         $params{ip}||='0.0.0.0';            # INET_ANY
-        $params{port}||=603;                # POE! (almost :)
+        $params{port} = 603                 # POE! (almost :)
+                    unless defined $params{port};
     }
+    $params{protocol} ||= 'IKC0';
 
-    create_ikc_responder();
-    POE::Session->create(
+    # Make sure one is available
+    POE::Component::IKC::Responder->spawn();
+    my $session = POE::Session->create(
                     package_states => [ 
                         $params{package} =>
                         [qw(
-                            _start _stop error
+                            _start _stop error _child
                             accept fork retry waste_time
                             babysit rogues shutdown
                             sig_CHLD sig_INT sig_USR2 sig_USR1 sig_TERM
@@ -68,66 +74,34 @@ sub create_ikc_server
                     ],
                     args=>[\%params],
                   );
+    my $heap = $session->get_heap;
+    return $heap->{wheel_port};
 }
 
-sub spawn
+sub create_ikc_server
 {
-    my($package, %params)=@_;
-    $params{package}=$package;
-    create_ikc_server(%params);
+    my( %params )=@_;
+    $params{package} ||= __PACKAGE__;
+    carp "create_ikc_server is DEPRECATED.  Please use $params{package}->spawn instead";
+    return $params{package}->spawn( %params );
 }
 
 #----------------------------------------------------
-# NOTE : THIS IS POORLY BEHAVED CODE
 sub _select_define
 {
     my($heap, $on)=@_;
+    return unless $heap->{wheel};
     $on||=0;
 
     DEBUG and 
-        warn "_select_define (on=$on)";
-
-    if($POE::VERSION >= 0.24) {
-        if($on) {
-            $heap->{wheel}->resume_accept
-        }
-        else {
-            $heap->{wheel}->pause_accept
-        }
-        return;
-    }
-
-    # NOTE : FOR OLDER VERSIONS OF POE
-    # THIS IS POORLY BEHAVED CODE
-    my $state;
-    my $err="possible redefinition of POE internals";
-    $err="failure to bind to port" if $POE::VERSION <= 0.1005;
+        warn "_select_define  (on=$on)";
 
     if($on) {
-        if(ref $heap->{wheel} eq 'HASH') {
-            $state=$heap->{wheel}->{'state_accept'};
-        } else {
-            $state=$heap->{wheel}->[5];
-        }
-
-        unless(defined $state) {
-            die "No 'state_accept' in $heap->{wheel}, $err.\n";
-        }
-    }
-
-    my $c=0;
-    if(ref $heap->{wheel} eq 'HASH') {
-        foreach my $hndl (qw(socket_handle)) {
-            next unless defined $heap->{wheel}->{$hndl};
-            $poe_kernel->select_read($heap->{wheel}->{$hndl}, $state);
-            $c++;
-        }
+        $heap->{wheel}->resume_accept
     }
     else {
-        $poe_kernel->select_read($heap->{wheel}->[0], $state);
-        $c++;
+        $heap->{wheel}->pause_accept
     }
-    die "No socket_handle in $heap->{wheel}, $err.\n" unless $c;
     return;
 }
 
@@ -142,6 +116,38 @@ sub _delete_wheel
     return;
 }
 
+#----------------------------------------------------
+#
+sub _concurrency_up
+{
+    my( $heap ) = @_;
+    $heap->{concur_connections}++;
+    DEBUG and 
+        warn "$$: $heap->{concur_connections} concurrent connections (max $heap->{concurrency})";
+    return unless $heap->{concurrency} > 0;
+    if( $heap->{concur_connections} >= $heap->{concurrency} ) {
+        DEBUG and 
+            warn "$$: Blocking more concurrency";
+        $heap->{blocked} = 1;
+        _select_define( $heap, 0 );
+    }
+}
+
+
+sub _concurrency_down
+{
+    my( $heap ) = @_;
+    $heap->{concur_connections}--;
+    DEBUG and 
+        warn "$$: $heap->{concur_connections} concurrent connections";
+    return unless $heap->{concurrency} > 0;
+    if( $heap->{concur_connections} < $heap->{concurrency} 
+        and delete $heap->{blocked} ) {
+        DEBUG and 
+            warn "$$: Unblocking concurrency";
+        _select_define( $heap, 1 );
+    }
+}
 
 #----------------------------------------------------
 # Delete all delays
@@ -164,6 +170,11 @@ sub _start
 {
     my($heap, $params, $kernel) = @_[HEAP, ARG0, KERNEL];
 
+    my $ret;
+
+    # This shouldn't be necessary
+    POE::Component::IKC::Responder->spawn;
+
     # monitor for shutdown events.
     # this is the best way to get IKC::Responder to tell us about the
     # shutdown
@@ -184,19 +195,32 @@ sub _start
     }
     else {
         $alias="$params->{ip}:$params->{port}";
-        $wheel_p{BindPort}= $params->{port};
+        $wheel_p{BindPort} = $params->{port};
         $wheel_p{BindAddress}= $params->{ip};
     }
     DEBUG && warn "$$: Server starting $alias.\n";
 
-    # +GC
-    $kernel->alias_set("IKC Server $alias");
 
-                                        # create a socket factory
-    $heap->{wheel_address}=$alias;
-    $heap->{wheel} = new POE::Wheel::SocketFactory (%wheel_p);
     $heap->{name}=$params->{name};
     $heap->{kernel_aliases}=$params->{aliases};
+    $heap->{concurrency}=$params->{concurrency} || 0;
+    $heap->{protocol}=$params->{protocol};
+    $heap->{on_error}=$params->{on_error} if $params->{on_error};
+                                        # create a socket factory
+    $heap->{wheel} = new POE::Wheel::SocketFactory (%wheel_p);
+    if( $heap->{wheel} and not $params->{unix} and not $params->{port} ) {
+        $heap->{wheel_port} = 
+                $ret = ( sockaddr_in( $heap->{wheel}->getsockname() ) )[0];
+        $alias="$params->{ip}:$ret";
+        DEBUG && 
+                warn "$$: Server listening on $alias.\n";
+    }
+    $heap->{wheel_address}=$alias;
+
+    $heap->{connections} = 0;
+
+    # +GC
+    $kernel->alias_set("IKC Server $alias");
 
     # set up local names for kernel
     my @names=($heap->{name});
@@ -210,7 +234,16 @@ sub _start
 
     $kernel->post(IKC=>'register_local', \@names);
 
-    return unless $params->{processes};
+    # pre-load the default serialisers
+    foreach my $ft ( qw(Storable FreezeThaw POE::Component::IKC::Freezer) ) {
+        eval {  local $SIG{__WARN__} = sub {1}; 
+                local $SIG{__DIE__} = 'DEFAULT';
+                POE::Filter::Reference->new( $ft );
+             };
+        warn "$ft: $@" if DEBUG and $@;
+    }
+
+    return $ret unless $params->{processes};
 
     # Delete the SocketFactory's read select in the parent
     # We don't ever want the parent to accept a connection
@@ -249,7 +282,46 @@ sub _start
         };
         $kernel->yield('babysit');
     }
-    return;
+    return $ret;
+}
+
+#------------------------------------------------------------------------------
+sub _child
+{
+    my( $heap, $kernel, $op, $child, $ret ) = 
+                                @_[ HEAP, KERNEL, ARG0, ARG1, ARG2 ];
+    $ret ||= '';
+    DEBUG and 
+        warn "$$: _child op=$op child=$child ret=$ret";
+    unless( $ret eq "channel-$child" ) {
+        if( $op eq 'create' ) {
+            DEBUG and 
+                warn "$$: Detatching child session $child";
+            $kernel->detach_child( $child );
+        }
+        return;
+    }
+    if( $op eq 'lose' ) {
+        DB::disable_profile() if $INC{'Devel/NYTProf.pm'};
+        $heap->{child_sessions}--;
+        if( $heap->{child_sessions} > 0 ) {
+            DEBUG and warn "$$: still have a child session";
+        }
+        _concurrency_down($heap);
+    }
+    else {
+        $heap->{child_sessions}++;
+        return;
+    }
+    unless( $heap->{wheel} ) {  # no wheel == GAME OVER
+        ( DEBUG and not $INC{'Test/More.pm'} ) and
+            warn "$$: }}}}}}}}}}}}}}} Game over\n";
+        # XXX: Using shutdown is a stop-gap measure.  Maybe the daemon
+        # wants to stay alive even if IKC was shutdown...
+        # XXX: more to the point, maybe there are still requests that are
+        # hanging around !
+        $kernel->call( IKC => 'shutdown' );
+    }
 }
 
 #------------------------------------------------------------------------------
@@ -260,7 +332,7 @@ sub waste_time
     return if $heap->{'is a child'};
 
     unless($heap->{'been told we are parent'}) {
-        warn "$$: Telling everyone we are the parent\n";
+        $heap->{verbose} and warn "$$: Telling everyone we are the parent\n";
         $heap->{'been told we are parent'}=1;
         $kernel->signal($kernel, '__parent');
     }
@@ -312,13 +384,13 @@ sub babysit
                 $time=eval{$table{$pid}->utime + $table{$pid}->stime};
                 warn $@ if $@;
                 # utime and stime are Linux-only :(
-                $time /= 1_000_000 if $time;    # in micro-seconds
+                $time /= 1_000_000 if $time;    # micro-seconds -> seconds
 
-                if($time and $time > 600) { # arbitrary limit of 10 minutes
+                if($time and $time > 1200) {    # arbitrary limit of 20 minutes
                     $rogues{$pid}=$table{$pid};
                         warn "$$: $pid has gone rogue, time=$time s\n";
                 } else {
-                    # DEBUG and 
+                    DEBUG and 
                         warn "$$: child $pid has utime+stime=$time s\n"
                             if $time > 1;
                     $ok{$pid}=1;
@@ -462,7 +534,7 @@ sub shutdown
     _delete_delays();               # get it OVER with
 
     # -GC
-    $kernel->alias_remove("IKC Server $heap->{wheel_address}");
+    # $kernel->alias_remove("IKC Server $heap->{wheel_address}");
     $heap->{'die'}=1;               # prevent race conditions
 }
 
@@ -474,21 +546,23 @@ sub shutdown
 sub error
 {
     my ($heap, $operation, $errnum, $errstr) = @_[HEAP, ARG0, ARG1, ARG2];
-    warn __PACKAGE__, " $$: encountered $operation error $errnum: $errstr\n";
+
+
+    DEBUG and
+        warn __PACKAGE__, " $$: encountered $operation error $errnum: $errstr\n";
+
     my $ignore;
     if($errnum==EADDRINUSE) {       # EADDRINUSE
-        warn "$$: IKC Address $heap->{wheel_address} in use\n";
         $heap->{'die'}=1;
         _delete_wheel( $heap );
-        $ignore=1;
+        $ignore = 0;
     } elsif($errnum==WSAEAFNOSUPPORT) {
         # Address family not supported by protocol family.
         # we get this error, yet nothing bad happens... oh well
         $ignore=1;
     }
     unless($ignore) {
-        # TODO : post to monitors
-        warn __PACKAGE__, " $$: encountered $operation error $errnum: $errstr\n";
+        POE::Component::IKC::Util::monitor_error( $heap, $operation, $errnum, $errstr );
     }
 }
 
@@ -501,6 +575,7 @@ sub accept
     my ($heap, $kernel, $handle, $peer_host, $peer_port) = 
             @_[HEAP, KERNEL, ARG0, ARG1, ARG2];
 
+    T->start( 'IKC' );
     if(DEBUG) {
         if($peer_port) {        
             warn "$$: Server connection from ", inet_ntoa($peer_host), 
@@ -515,33 +590,37 @@ sub accept
     }
     if($heap->{children} and not $heap->{'is a child'}) {
         warn "$$: Parent process received a connection: THIS SUCKS\n";
-        _select_define($heap, 0);
         return;
     }
 
+    DB::enable_profile() if $INC{'Devel/NYTProf.pm'};
+
     DEBUG and warn "$$: Server kernel_aliases=", join ',', @{$heap->{kernel_aliases}||[]};
 
                                         # give the connection to a channel
     POE::Component::IKC::Channel->spawn(
-                handle=>$handle, name=>$heap->{name},
-                unix=>$heap->{unix}, aliases=>[@{$heap->{kernel_aliases}||[]}]);
+                handle=>$handle, 
+                name=>$heap->{name},
+                unix=>$heap->{unix}, 
+                aliases=>[@{$heap->{kernel_aliases}||[]}],
+                protocol=>$heap->{protocol},
+                on_error=>$heap->{on_error}
+            );
+
+    _concurrency_up($heap);
         
     return unless $heap->{children};
 
-    if ($heap->{'is a child'}) {
+    if (--$heap->{connections} < 1) {
+        DEBUG and 
+                warn "$$: {{{{{{{{{{{{{{{ Game over\n";
+        $kernel->delay('waste_time');
+        _delete_wheel( $heap );
+        $::TRACE_REFCNT = 1;
 
-        if (--$heap->{connections} < 1) {
-            # DEBUG and 
-                warn "$$: ************* Game over\n";
-            $kernel->delay('waste_time');
-            _delete_wheel( $heap );
-
-        } else {
-            # DEBUG and 
-                warn "$$: $heap->{connections} connections left\n";
-        }
     } else {
-        warn "$$: Master client got a connect!  this sucks!\n";
+        DEBUG and 
+                warn "$$: $heap->{connections} connections left\n";
     }
 }
 
@@ -573,9 +652,8 @@ sub fork
             $heap->{'failed forks'}++;
             $kernel->delay('retry', 1);
         }
-                                        # fail permanently, if fatal
-        else {
-            warn "Can't fork: $!\n";
+        else {                          # fail permanently, if fatal
+            POE::Component::IKC::Util::monitor_error( $heap, 'fork', 0+$1, "$!" );
             $kernel->yield('_stop');
         }
         return;
@@ -595,7 +673,12 @@ sub fork
 
         # This resets some kernel data that was preventing the child process's
         # kernel from becoming IDLE
-        $kernel->_data_sig_initialize;
+        if( $kernel->can( 'has_forked' ) ) {
+            $kernel->has_forked;
+        }
+        else {
+            $kernel->_data_sig_initialize;
+        }
 
         # Clean out stuff that the parent needs but not the children
         $heap->{'is a child'}   = 1;        # don't allow fork
@@ -693,7 +776,7 @@ sub sig_INT
     return 0 if $heap->{"is a child"};
 
     if($heap->{children}) {
-        warn "$$ SIGINT\n";
+        $heap->{verbose} and warn "$$ SIGINT\n";
         $heap->{'die'}=1;
         # kill all events
         _delete_delays();               # get it OVER with
@@ -714,7 +797,7 @@ sub sig_TERM
     my ($kernel, $heap, $signal, $pid, $status) =
                 @_[KERNEL, HEAP, ARG0, ARG1, ARG2];
 
-    warn "$$ SIGTERM\n";
+    $heap->{verbose} and warn "$$ SIGTERM\n";
     $heap->{'die'}=1;
 
     _delete_wheel( $heap );
@@ -767,20 +850,22 @@ sub __peek
 {
     my($verbose)=@_;
     eval {
-        require POE::API::Peek;
         require POE::Component::Daemon;
     };
+    unless( $@ ) {
+        my $ret = Daemon->peek( $verbose );
+        $ret =~ s/\n/\n$$: /g;
+        warn "$$: $ret";
+        return 1;
+    }
+
+    eval {
+        require POE::API::Peek;
+    };
     if($@) {
         DEBUG and warn "Failed to load POE::API::Peek: $@";
         return;
     }
-    my $ret = Daemon->peek( $verbose );
-
-    $ret =~ s/\n/\n$$: /g;
-    warn "$$: $ret";
-    return 1;
-
-
     my $api=POE::API::Peek->new();
     my @queue = $api->event_queue_dump();
     
@@ -809,7 +894,7 @@ sub __peek
         $ret.="Keepalive " unless $verbose;
         $ret.="Sessions: \n";
         my $ses;
-        foreach my $session ($api->session_list) {  
+        foreach my $session ( sort { $a->ID <=> $b->ID } $api->session_list) {  
             my $ref=0;
             $ses='';
 
@@ -844,7 +929,7 @@ sub __peek
                 $ref += $q1;
             }
 
-            my $q1 = $events->{ $session->ID }{destination};
+            $q1 = $events->{ $session->ID }{destination};
             if( $q1 ) {
                 $ret.="\t\tEvent destination count: $q1 (Stay alive)\n";
                 $ref += $q1;
@@ -904,10 +989,10 @@ POE::Component::IKC::Server - POE Inter-kernel Communication server
 
     use POE;
     use POE::Component::IKC::Server;
-    create_ikc_server(
+    POE::Component::IKC::Server->spawn(
         ip=>$ip, 
         port=>$port,
-        name=>'Server',);
+        name=>'Server');
     ...
     $poe_kernel->run();
 
@@ -922,10 +1007,16 @@ identical.
 
 =head2 C<create_ikc_server>
 
-This function initiates all the work of building the IKC server.  
+Deprecated.  Use L<POE::Component::IKC::Server/spawn>.
+
+=head1 CLASS METHODS
+
+=head2 C<spawn>
+
+This methods initiates all the work of building the IKC server. 
 Parameters are :
 
-=over 3
+=over 4
 
 =item C<ip>
 
@@ -934,7 +1025,14 @@ Address to listen on.  Can be a doted-quad ('127.0.0.1') or a host name
 
 =item C<port>
 
-Port to listen on.  Can be numeric (80) or a service ('http').
+Port to listen on.  Can be numeric (80) or a service ('http').  If
+undefined, will default to 603.  
+If you set the port to 0, a random port will be chosen and C<spawn> will
+return the port number.
+
+    my $port = POE::Component::IKC::Server->spawn( port => 0 );
+    warn "Listeing on port $port";
+
 
 =item C<unix>
 
@@ -976,8 +1074,45 @@ connections are serviced concurrently, because there's no way to know when
 we have finished a request.  Defaults to 1 (ie, one connection per
 child).
 
+=item C<concurrency>
+
+Number of simultaneous connected clients allowed.
+Defaults to 0 (unlimited).  
+
+Note that this is per-IKC::Server instance;  if you have several ways of
+connecting to a give IKC server (for example, both an TCP/IP port and unix
+pipe), they will not share the conncurrent connection count.
+
+=item C<protocol>
+
+Which IKC negociation protocol to use.  The original protocol (C<IKC>) had a
+slow synchronous handshake.  The new protocol (C<IKC0>) sends all the
+handshake information at once.  IKC0 will degrade gracefully to IKC, if the
+client and server don't match.
+
+Default is IKC0.
+
+=item C<on_error>
+
+Coderef that is called for all errors. You could use this to monitor for
+problems when forking children or opening the socket.  Parameters are
+C<$operation, $errnum and $errstr>, which correspond to
+POE::Wheel::SocketFactory's FailureEvent, which q.v.
+
+However, IKC/monitor provides a more powerful mechanism for detecting
+errors.  See L<POE::Component::IKC::Responder>.  
+
+Note, also, that the coderef will be executed from within an IKC session,
+NOT within your own session.  This means that things like
+$poe_kernel->delay_set() won't do what you think they should.
+
+
 =back
 
+C<POE::Component::IKC::Server::spawn> returns C<undef()>, unless you specify
+a L</port>=0, in which case, C<spawn> returns the port that was chosen.
+
+
 =head1 EVENTS
 
 =head2 shutdown
@@ -998,7 +1133,7 @@ Philip Gwyn, <perl-ikc at pied.nu>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright 1999-2008 by Philip Gwyn.  All rights reserved.
+Copyright 1999-2014 by Philip Gwyn.  All rights reserved.
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself.
